#include <vtk.h>
#include <cave_ogl.h>
#include <iostream.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <math.h>
#include <vtkCAVERenderWindow.h>
#include <vtkCAVERenderer.h>

#ifndef DTOR
#define DTOR(r) r*(M_PI/180.0)
#endif

// The vtkCAVE variables
vtkCAVERenderWindow* caveRenderWindow;
vtkCAVERenderer* caveRenderer;

// Note vtk objects can NOT be put into
// shared memory, hence the reason for global
// distributed memory objects.
vtkActor* cylinderActor;
vtkMatrix4x4* cylinderMatrix;
vtkTransform* caveTransform;
vtkPlane* plane;
vtkActor* cutActor;
float *segment, *offset;

// Function prototypes
void draw();
void frameUpdate();
void createCombustor(vtkPLOT3DReader*);
void getWandMatrix(vtkMatrix4x4*);
void initVTK(int, char**);
void navTranslate(float);
void CAVENavRotateY();
void navigate();


// This is the draw function that gets passed to the CAVELib
// All rendering should be done in this function between the 
// glPushMatrix() and glPopMatrix() calls.
void
draw()
{
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    glClearColor(0,0,0,0);

    glPushMatrix();

       // Get the cave Navigation Transformation.  What allows the user
       // to move in the CAVE(with wand after a little help from us in
       // the frame function, navigate function call)
       CAVENavTransform();
       caveRenderWindow->Render();

    glPopMatrix();
}


// This frame function gets called once per frame
// and handles user interactions.
void
frameUpdate(){

    float or[3], *origin;
     
    if(CAVEBUTTON1){
       // If the user presses the left button on the wand
       // the angle of the wand will determine the position of
       // the cutting plane within the cumbuster.  -90 is the
       // bottom of the cumbuster, 0 is center, 90 is the top.
        
       // retrieve the wand's orientation
       CAVEGetOrientation(CAVE_WAND_NAV,or);
     
       // Looking at the X angle
       or[0] += 90;
       if(or[0] < 0) or[0] = 0;
       if(or[0] > 180) or[0] = 180;

       // Get the where the plane is now
       origin = plane->GetOrigin();
       // move the plane to the new position
       plane->SetOrigin(origin[0], origin[1], *offset + *segment*or[0]);
    }


    if(CAVEBUTTON3){  // allow resizing of the cumbustion chamber
 
       if(CAVE_JOYSTICK_Y > 0.5)
       {
          caveTransform->Translate(0.0, -5.0, 0.0);
          caveTransform->Scale(1.025, 1.025, 1.025);
          caveTransform->Translate(0.0, 5.0, 0.0);
       }if(CAVE_JOYSTICK_Y < -0.5)
       {
          caveTransform->Translate(0.0, -5.0, 0.0);
          caveTransform->Scale(0.975, 0.975, 0.975);
          caveTransform->Translate(0.0, 5.0, 0.0);
       }
        
    }
    else
       navigate(); // process navigation on the wand


    // Get the transformation matrix of the wand and set the cylinder
    // to be that same transformation
    getWandMatrix(cylinderMatrix);
    cylinderActor->SetUserMatrix(cylinderMatrix);

}


// Initialize the vtk objects
void
initVTK(int argc, char** argv)
{
    float pos[] = {-5, 0, 5};

    // Create the necessary vtkCAVE objects
    caveRenderer = vtkCAVERenderer::New();
    caveRenderWindow = vtkCAVERenderWindow::New();

    caveRenderWindow->AddRenderer(caveRenderer);
    //caveRenderer->TwoSidedLightingOn();

    // Create some lights
    vtkLight* light1 = vtkLight::New();
    vtkLight* light2 = vtkLight::New();
    vtkLight* light3 = vtkLight::New();
    vtkLight* light4 = vtkLight::New();

    light1->SetPosition(pos[0], pos[1], pos[2]);
    light1->SetConeAngle(180);
    light1->SetColor(1.0, 1.0, 1.0);

    pos[0] = 5;
    light2->SetPosition(pos[0], pos[1], pos[2]);
    light2->SetConeAngle(180);
    light2->SetColor(1.0, 1.0, 1.0);

    pos[0] = 0;
    pos[1] = 10;
    pos[2] = 3;
    light3->SetPosition(pos[0],pos[1],pos[2]);
    light3->SetConeAngle(180);
    light3->SetColor(1,1,1);

    pos[0] = 0;
    pos[1] = 0;
    pos[2] = 0;
    light4->SetPosition(pos[0],pos[1],pos[2]);
    light4->SetConeAngle(180);
    light4->SetFocalPoint(0.0, 5.0, 0.0);
    light4->SetColor(1,1,1);

    caveRenderer->AddLight(light1);
    caveRenderer->AddLight(light2);
    caveRenderer->AddLight(light3);
    caveRenderer->AddLight(light4);

    // Create the reader
    vtkPLOT3DReader *reader = vtkPLOT3DReader::New();

    if(argc > 1){ // there is a arguement telling the 
          // directory where the .bin files are
       char *dataSetLocation = new char[80];
       sprintf(dataSetLocation, "%s/combxyz.bin", argv[1]);
       reader->SetXYZFileName(dataSetLocation);
       sprintf(dataSetLocation, "%s/combq.bin", argv[1]);
       reader->SetQFileName(dataSetLocation);
       delete(dataSetLocation);
    }else{ // no arguement assuming current directory
       reader->SetXYZFileName("combxyz.bin");
       reader->SetQFileName("combq.bin");
    }
 
    reader->SetScalarFunctionNumber(100);
    reader->SetVectorFunctionNumber(202);
    reader->Update();

    // Create the Cumbustor and cutting Plane
    createCombustor(reader);
}

// create the combustion chamber outline and the cutting plane
void
createCombustor(vtkPLOT3DReader* reader)
{
    float *bounds, *center, range;

    center = reader->GetOutput()->GetCenter();
    caveTransform = vtkTransform::New();
      // Translate to center of CAVE
      caveTransform->Translate(0.0, 5.0, 0.0);
      caveTransform->Scale(0.125, 0.125, 0.125);
      caveTransform->RotateX(-90);
      caveTransform->RotateY(17.43);
      // translate to the center(ie 0,0,0) for rotation
      caveTransform->Translate(-center[0], -center[1], -center[2]);
    
    // Create a cylinder that will represent the ray from
    // the front of the wand
    vtkCylinderSource* cylinder = vtkCylinderSource::New();
    cylinder->SetRadius(0.05);
    cylinder->SetHeight(5.0);

    // Map the objects to graphics primitives
    vtkDataSetMapper* cylinderMapper = vtkDataSetMapper::New();
    cylinderMapper->SetInput(cylinder->GetOutput());

    cylinderActor = vtkActor::New();
    cylinderActor->SetMapper(cylinderMapper);
    cylinderActor->SetPosition(0,0,-2.5);
    cylinderActor->RotateX(90.0);
    cylinderActor->PickableOff();

    // Create a global matrix object to be used
    // in getting the orientation of the wand and
    // setting the ray cylinder to align with the wand
    cylinderMatrix = vtkMatrix4x4::New();

    // add the wand pointer
    caveRenderer->AddActor(cylinderActor);

//
// Outline of the Cumbustion Chamber
//
  vtkStructuredGridOutlineFilter *outlineF = vtkStructuredGridOutlineFilter::New();
    outlineF->SetInput(reader->GetOutput());
  vtkPolyDataMapper *outlineMapper = vtkPolyDataMapper::New();
    outlineMapper->SetInput(outlineF->GetOutput());
  vtkActor *outline = vtkActor::New();
    outline->SetMapper(outlineMapper);
    outline->GetProperty()->SetColor(1,1,1);
    outline->SetUserMatrix(caveTransform->GetMatrixPointer());
//
// create the cutting plane
//
    plane = vtkPlane::New();
       plane->SetOrigin(reader->GetOutput()->GetCenter());
       plane->SetNormal(-0.287, 0.0, 0.9579);
    vtkCutter *planeCut = vtkCutter::New();
       planeCut->SetInput(reader->GetOutput());
       planeCut->SetCutFunction(plane);
    vtkDataSetMapper *cutMapper = vtkDataSetMapper::New();
       cutMapper->SetInput(planeCut->GetOutput());
       cutMapper->SetScalarRange(\
            reader->GetOutput()->GetPointData()->GetScalars()->GetRange());
    cutActor = vtkActor::New();
       cutActor->SetMapper(cutMapper);
       cutActor->SetUserMatrix(caveTransform->GetMatrixPointer());


    caveRenderer->AddActor(outline);   
    caveRenderer->AddActor(cutActor);   

 //cutActor->SetBackfaceProperty(cutActor->GetProperty());

    // set up for movement of the cutting plane
    segment = (float*) CAVEMalloc(sizeof(float));
    offset = (float*) CAVEMalloc(sizeof(float));

    bounds = reader->GetOutput()->GetBounds();
    range = bounds[5] - bounds[4];
    *segment = range / 180;
    *offset = bounds[4];

    // set up for resizing of the cumbustor
    caveTransform->PostMultiply();
}

void
main(int argc, char **argv)
{

    CAVEConfigure(&argc,argv,NULL);
    CAVEInit();
    CAVEInitApplication((CAVECALLBACK)initVTK, 2, argc, argv);
    CAVEFrameFunction(frameUpdate, 0);
    CAVEDisplay(draw, 0);
    
    while(!CAVEgetbutton(CAVE_ESCKEY))
    {
    }

    // free up the memory we allocated
    CAVEFree(segment);
    CAVEFree(offset);
    
    CAVEExit();

    caveRenderer->Delete();
    caveRenderWindow->Delete();
}

// From the wands position and orientation determine
// its transformation matrix.
void
getWandMatrix(vtkMatrix4x4* m)
{
    vtkTransform *transform;
    float or[3],pos[3];

    transform = vtkTransform::New();

    CAVEGetPosition(CAVE_WAND_NAV,pos);
    CAVEGetOrientation(CAVE_WAND_NAV,or);

    // note the order of rotation.  This is very important
    transform->Translate(pos[0], pos[1], pos[2]);
    transform->RotateY(or[1]);
    transform->RotateX(or[0]);
    transform->RotateZ(or[2]);

    m->DeepCopy(transform->GetMatrixPointer());
    transform->Delete();

}

// Handles the translation component of navigation
// based on the orientation of the wand
void
navTranslate(float dz)
{

    float wandOrn[3];
    float wx, wz;

    CAVEGetOrientation(CAVE_WAND,wandOrn);
    wx = dz * sinf(DTOR(wandOrn[1]))*5;
    wz = dz * cosf(DTOR(wandOrn[1]))*5;
    CAVENavTranslate(wx, 0, wz);

}


// Handles the rotation component of navigation
// based on the amount the joystick is pressed
void
CAVENavRotateY()
{
    float azi;

    azi = 1.0;

    if(CAVE_JOYSTICK_X > 0.5)
        CAVENavRot(-azi, 'y');

    if(CAVE_JOYSTICK_X < -0.5)
        CAVENavRot(azi, 'y');

}


// The basic navigation function. It calls the
// auxillary navigation functions.
void
navigate()
{

    float dz;

    if(CAVE_JOYSTICK_Y > 0.5)
    {
        dz = -.1;
        navTranslate(dz);
    }
    if(CAVE_JOYSTICK_Y < -0.5)
    {
        dz = .1;
        navTranslate(dz);
    }
    CAVENavRotateY();
}

