#include <vtk.h>
#include <cave_ogl.h>
#include <iostream.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <vtkCAVERenderWindow.h>
#include <vtkCAVERenderer.h>
#include <vtkMarchingCubes.h>
#include <vtkStructuredPointsToPolyDataFilter.h>


// Some data constants
const int DIMS = 256;
const int XDIM = 64;
const int YDIM = 64;
const int SLICES = 93;
const float XSPACING = 0.8;
const float YSPACING = 0.8;
const float ZSPACING = 1.5;


// The vtkCAVE variables
vtkCAVERenderWindow* caveRenderWindow;
vtkCAVERenderer* caveRenderer;


// Create a global vtkProperty object so that
// property changes can be done in the frame update
// function. Note vtk objects can NOT be put into
// shared memory, hence the reason for global
// distributed memory objects.
vtkProperty* prop;


// Function prototypes
void draw();
void frameUpdate();
void createIsosurface(vtkVolume16Reader*);
void initVTK(int, char**);

// This is the draw function that gets passed to the CAVELib
// All rendering should be done in this fuction between the 
// glPushMatrix() and glPopMatrix() calls.
void
draw()
{
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    glClearColor(0,0,0,0);

    glPushMatrix();
    caveRenderWindow->Render();
    glPopMatrix();
}


// This frame function gets called once per frame
// and handles user interactions.
void
frameUpdate()
{
    int change1;

    change1 = CAVEButtonChange(1);
  
    // While the user has the left button depressed
    // show the frame outline of the isosurface.
    if(change1 == 1)
	prop->SetRepresentationToWireframe();
    else if(change1 == -1)
	prop->SetRepresentationToSurface();
}


// Initialize the vtk objects
void
initVTK(int argc, char** argv)
{
    float pos[3] = {-5, 5, 0};

    // Create the necessary vtkCAVE objects
    caveRenderer = vtkCAVERenderer::New();
    caveRenderWindow = vtkCAVERenderWindow::New();

    // Create some lights
    vtkLight* light1 = vtkLight::New();
    vtkLight* light2 = vtkLight::New();
    vtkLight* light3 = vtkLight::New();

    light1->SetPosition(pos[0], pos[1], pos[2]);
    light1->SetColor(1.0, 1.0, 1.0);
    pos[0] = 5;

    light2->SetPosition(pos[0], pos[1], pos[2]);
    light2->SetColor(1.0, 1.0, 1.0);
    pos[0] = 0;
    pos[1] = 10;
    pos[2] = 3;

    light3->SetPosition(pos[0],pos[1],pos[2]);
    light3->SetColor(1,1,1);

    // Add the lights to the caveRenderer
    caveRenderer->AddLight(light1);
    caveRenderer->AddLight(light2);
    caveRenderer->AddLight(light3);
    caveRenderWindow->AddRenderer(caveRenderer);

    // Create the volume reader
    vtkVolume16Reader *v16 = vtkVolume16Reader::New();
    v16->SetDataDimensions(XDIM,YDIM);
    v16->SetDataByteOrderToLittleEndian();

    // Get the location of the data files
    char *datasetlocation = new char[80];
    if(argc > 1)
       sprintf(datasetlocation, "%s", argv[1]); 
    else
       sprintf(datasetlocation, "./");

    // Specify the location of the head data
    sprintf(datasetlocation,"%s/headsq/quarter",datasetlocation);
    v16->SetFilePrefix (datasetlocation);
    v16->SetImageRange(1,SLICES);
    v16->SetDataSpacing (XSPACING*(DIMS/XDIM), YSPACING*(DIMS/YDIM), ZSPACING);

    // Create the isosurface
    createIsosurface(v16);
}


// Create a basic isosurface
void
createIsosurface(vtkVolume16Reader* v16)
{

    // Set up the marching cubes
    vtkMarchingCubes *mcube = vtkMarchingCubes::New();
    mcube->SetInput(v16->GetOutput());
    mcube->ComputeScalarsOff();
    mcube->ComputeGradientsOff();
    mcube->ComputeNormalsOff();
    mcube->SetValue(0, 500);

    // Create some normals
    vtkPolyDataNormals* normals = vtkPolyDataNormals::New();
    normals->SetFeatureAngle(60);
    normals->SetInput(mcube->GetOutput());

    // Use the vtkStripper, it greatly increases performance
    vtkStripper *boneStripper = vtkStripper::New();
    boneStripper->SetInput(normals->GetOutput());

    // Map the data to polygons
    vtkPolyDataMapper *boneMapper = vtkPolyDataMapper::New();
    boneMapper->SetInput(boneStripper->GetOutput());
    boneMapper->ScalarVisibilityOff();

    // Create the actor
    vtkActor* contourSurface = vtkActor::New();
    contourSurface->SetMapper(boneMapper);

    // Create the global property object
    prop = vtkProperty::New();

    contourSurface->SetProperty(prop);

    contourSurface->GetProperty()->SetDiffuseColor(1, 1, .9412);
    contourSurface->GetProperty()->SetSpecular(.3);
    contourSurface->GetProperty()->SetSpecularPower(20);

    // Transform the object to be viewable in the CAVE.
    vtkTransform *myTrans = vtkTransform::New();
    myTrans->Translate(-1, 6.0, -2.5);
    myTrans->RotateX(90.0);
    myTrans->Scale(2.0/(XSPACING*DIMS),
		   2.0/(XSPACING*DIMS),
		   2.0/(XSPACING*DIMS));
    
    contourSurface->SetUserMatrix(&myTrans->GetMatrix());

    // Add the actor to the cave renderer
    caveRenderer->AddActor(contourSurface);   
}

void
main(int argc, char **argv)
{

    CAVEConfigure(&argc,argv,NULL);
    CAVEInit();
    CAVEInitApplication((CAVECALLBACK)initVTK, 2, argc, argv);
    CAVEFrameFunction(frameUpdate, 0);
    CAVEDisplay(draw, 0);
    
    while(!CAVEgetbutton(CAVE_ESCKEY))
    {
	sginap(10);
    }
    
    CAVEExit();
}
