#include <vtk.h>
#include <cave_ogl.h>
#include <iostream.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <math.h>
#include <vtkCAVERenderWindow.h>
#include <vtkCAVERenderer.h>
#include <vtkCAVEPicker.h>

#ifndef DTOR
#define DTOR(r) r*(M_PI/180.0)
#endif

// The vtkCAVE variables
vtkCAVERenderWindow* caveRenderWindow;
vtkCAVERenderer* caveRenderer;
vtkCAVEPicker* cavePicker;

// Note vtk objects can NOT be put into
// shared memory, hence the reason for global
// distributed memory objects.
vtkPlaneSource* rect;
vtkActor* wandActor;
vtkActor* lines;
vtkActor* rake;
vtkMatrix4x4* wandMatrix;
vtkTransform* cave2Actor;


// Function prototypes
void draw();
void frameUpdate();
void createRoom(vtkStructuredGridReader*);

void getWandMatrix(vtkMatrix4x4*);
void initVTK(int, char**);
void navTranslate(float);
void CAVENavRotateY();
void navigate();


// This is the draw function that gets passed to the CAVELib
// All rendering should be done in this fuction between the 
// glPushMatrix() and glPopMatrix() calls.
void
draw()
{
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    glClearColor(0,0,0,0);

    glPushMatrix();

    CAVENavTransform();
    caveRenderWindow->Render();
   
    glPopMatrix();
}


// This frame function gets called once per frame
// and handles user interactions.
void
frameUpdate(){

    float wandVector[3], wandPosition[3], worldPoint[4];
    float *rakeCenter, *pickPosition, distanceSqr;
    int button1Change;
    static bool quickMoveOn = false, highRes = false;
    static float distance;

    if(CAVEButtonChange(2) == 1)
    {
       quickMoveOn = !quickMoveOn;
     cerr << "Pressed button 2\n";
       if(!quickMoveOn && !caveRenderer->GetActors()->IsItemPresent(lines))
         caveRenderer->AddActor(lines);
    }
          

    if(CAVEButtonChange(3) == 1)
       if(highRes = !highRes)
          rect->SetResolution(5, 5);
       else
          rect->SetResolution(1, 1);
     
    // retrieve the wand's position and direction vector
    CAVEGetPosition(CAVE_WAND_NAV, wandPosition);
    CAVEGetVector(CAVE_WAND_FRONT_NAV, wandVector);

    // check to see if the wand is pointing at the rake
    if(cavePicker->Pick(caveRenderer, wandPosition, wandVector, 0, 10))
    {

rake->DebugOn();
	// change the color of the rake to indicate it is picked
        rake->GetProperty()->SetColor(.9, .1, .1);

	// see if button 1 has changed (depressed, released, or same) 
        // since last call to function
    	button1Change = CAVEButtonChange(1);

	if(quickMoveOn){
	   if(button1Change == 1) // button has been pressed
               // turn off the streamlines
               caveRenderer->RemoveActor(lines);
           if(button1Change == -1) // button has been released
               // turn the streamlines back on.
               caveRenderer->AddActor(lines);
        }

	if(CAVEBUTTON1){ // button 1 is depressed :(
		// Allow the user to move the rake

            // compenstate for 90 deg rotation of model before sending
            // the new normal for the plane
cerr << "Vector " << wandVector[0] << ","\
    << wandVector[1] << "," << wandVector[2] << endl;

            // need this to determine the distance from the wand's position
            // to the positon of the rake
            pickPosition = rake->GetCenter();

cerr << "PickP " << pickPosition[0] << ","\
    << pickPosition[1] << "," << pickPosition[2] << endl;

	    if(button1Change == 1) 
            {   // first time in, fix the distance from the wand to 
                // allow 3D movement of the rake
                distanceSqr = vtkMath::Distance2BetweenPoints(\
                                    pickPosition, wandPosition);
                distance = sqrt(distanceSqr);
            }

            // determine the intersection of the wand pointer and the rake
            worldPoint[0] = distance*wandVector[0] + wandPosition[0];
            worldPoint[1] = distance*wandVector[1] + wandPosition[1];
            worldPoint[2] = distance*wandVector[2] + wandPosition[2];
            worldPoint[3] = 1.0;

cerr << " worldPoint" << worldPoint[0] << ","\
    << worldPoint[1] << "," << worldPoint[2] << endl;

	    // do the inverse transform to translate
            // the world coordinates into mapper coordinates
            cave2Actor->SetPoint(worldPoint);
            rakeCenter = cave2Actor->GetPoint();

cerr << "distance " << distance << " centerp " << rakeCenter[0] << ","\
    << rakeCenter[1] << "," << rakeCenter[2] << endl;

            // change the location of the rake
    	    rect->SetCenter(rakeCenter[0], rakeCenter[1], \
                rakeCenter[2]);
            //rect->SetNormal(wandVector[0], -wandVector[2], wandVector[1]);
        }
rake->DebugOff();
    }else 
	// reset the color of the rake, it is no longer picked
        rake->GetProperty()->SetColor(.8, .6, .6);


    // Get the transformation matrix of the wand and set the cylinder
    // to be that same transformation
    getWandMatrix(wandMatrix);
    wandActor->SetUserMatrix(wandMatrix);

    // handle the user navigation done with the wand
    navigate();
}


// Initialize the vtk objects
void
initVTK(int argc, char** argv)
{
    float pos[] = {-5, 5, 0};

    // Create the necessary vtkCAVE objects
    caveRenderer = vtkCAVERenderer::New();
    caveRenderWindow = vtkCAVERenderWindow::New();
    cavePicker = vtkCAVEPicker::New();

    // add the renderer to the window object
    caveRenderWindow->AddRenderer(caveRenderer);

    // Using the Pick list so that the rake is the only thing that
    // can be picked
    cavePicker->InitializePickList();
    cavePicker->PickFromListOn();

    // Create some lights
    vtkLight* light1 = vtkLight::New();
    vtkLight* light2 = vtkLight::New();
    vtkLight* light3 = vtkLight::New();

    light1->SetPosition(pos[0], pos[1], pos[2]);
    light1->SetColor(1.0, 1.0, 1.0);
 
    pos[0] = 5;
    light2->SetPosition(pos[0], pos[1], pos[2]);
    light2->SetColor(1.0, 1.0, 1.0);

    pos[0] = 0;
    pos[1] = 10;
    pos[2] = 3;
    light3->SetPosition(pos[0],pos[1],pos[2]);
    light3->SetColor(1,1,1);

    caveRenderer->AddLight(light1);
    caveRenderer->AddLight(light2);
    caveRenderer->AddLight(light3);

    // Create the reader
    vtkStructuredGridReader *reader = vtkStructuredGridReader::New();

    // Get the location of the data files
    char *dataSetLocation = new char[80];

    // Specify the location of the kitchen data
    // if there is not arguement assume in local directory
    if(argc > 1)
        sprintf(dataSetLocation,"%s/kitchen.vtk",argv[1]);
    else
        sprintf(dataSetLocation,"kitchen.vtk");

fprintf(stderr, "%s", dataSetLocation);
    reader->SetFileName(dataSetLocation);
    reader->Update();
 
    delete(dataSetLocation);

    // Create the kitchen and and the streamlines
    createRoom(reader);
}

// create the kitchen and it's parts and the streamlines
void
createRoom(vtkStructuredGridReader* reader)
{

    float range[2];
    float maxVelocity, maxTime;

    // set variable needed for later calculations
    if ( reader->GetOutput()->GetPointData()->GetScalars() )
    {
       reader->GetOutput()->GetPointData()->GetScalars()->GetRange(range);
    }
    if ( reader->GetOutput()->GetPointData()->GetVectors() )
    {
       maxVelocity = reader->GetOutput()->GetPointData()->GetVectors()->GetMaxNorm();
       maxTime = 35.0*reader->GetOutput()->GetLength()/maxVelocity ; 
    }

    // Transformation from the orginal coordinate system to 
    // the CAVE.  Sized to match the CAVE walls
    vtkTransform *caveTransform = vtkTransform::New();
      // in CAVE y is up and z is depth vtk is opposite so rotate
      caveTransform->RotateX(-90);
      caveTransform->Translate(-5.0, -5.0, 0.0);
      caveTransform->Scale(10.0/7.0, 2.0, 4.0);
    
    // Inverse transform to go from world to mapper coordinates
    // used in the frame function to allow movement of the rake
    cave2Actor = vtkTransform::New();
      cave2Actor->SetMatrix(*caveTransform->GetMatrixPointer());
      cave2Actor->Inverse();
      cave2Actor->PostMultiply();

    // Create a cylinder that will represent the ray from
    // the front of the wand.  Makes picking easier for the user
    vtkCylinderSource* cylinder = vtkCylinderSource::New();
       cylinder->SetRadius(0.05);
       cylinder->SetHeight(5.0);

    // Map the objects to graphics primitives
    vtkDataSetMapper* cylinderMapper = vtkDataSetMapper::New();
       cylinderMapper->SetInput(cylinder->GetOutput());

    wandActor = vtkActor::New();
       wandActor->SetMapper(cylinderMapper);
       wandActor->SetPosition(0,0,-2.5);
       wandActor->RotateX(-90.0);
       wandActor->PickableOff(); // not needed since using 
          // a pick list.  VERY important if not using pick list

    // Create a global matrix object to be used
    // in getting the orientation of the wand and
    // setting the ray cylinder to align with the wand
    wandMatrix = vtkMatrix4x4::New();

    // register the 
    caveRenderer->AddActor(wandActor);

//
// The following 200 lines create the kitchen.  Taken from vtk example
// Kitchen.cxx addition of SetUserMatrix() to every Actor to make it
// use the caveTransform
// 

  vtkStructuredGridOutlineFilter *outlineF = vtkStructuredGridOutlineFilter::New();
    outlineF->SetInput(reader->GetOutput());
  vtkPolyDataMapper *outlineMapper = vtkPolyDataMapper::New();
    outlineMapper->SetInput(outlineF->GetOutput());
  vtkActor *outline = vtkActor::New();
    outline->SetMapper(outlineMapper);
    outline->GetProperty()->SetColor(1,1,1);
    outline->SetUserMatrix(caveTransform->GetMatrixPointer());

//
// Set up shaded surfaces (i.e., supporting geometry)
//
  vtkStructuredGridGeometryFilter *doorGeom = vtkStructuredGridGeometryFilter::New();
    doorGeom->SetInput(reader->GetOutput());
    doorGeom->SetExtent(27,27,14,18,0,11);
  vtkPolyDataMapper *mapDoor = vtkPolyDataMapper::New();
    mapDoor->SetInput(doorGeom->GetOutput());
    mapDoor->ScalarVisibilityOff();
  vtkActor *door = vtkActor::New();
    door->SetMapper(mapDoor);
    door->GetProperty()->SetColor(.59,.427,.3);
    door->SetUserMatrix(caveTransform->GetMatrixPointer());

  vtkStructuredGridGeometryFilter *window1Geom = vtkStructuredGridGeometryFilter::New();
    window1Geom->SetInput(reader->GetOutput());
    window1Geom->SetExtent(0,0,9,18,6,12);
  vtkPolyDataMapper *mapWindow1 = vtkPolyDataMapper::New();
    mapWindow1->SetInput(window1Geom->GetOutput());
    mapWindow1->ScalarVisibilityOff();
  vtkActor *window1 = vtkActor::New();
    window1->SetMapper(mapWindow1);
    window1->GetProperty()->SetColor(.3,.3,.5);
    window1->SetUserMatrix(caveTransform->GetMatrixPointer());

  vtkStructuredGridGeometryFilter *window2Geom = vtkStructuredGridGeometryFilter::New();
    window2Geom->SetInput(reader->GetOutput());
    window2Geom->SetExtent(5,12,23,23,6,12);
  vtkPolyDataMapper *mapWindow2 = vtkPolyDataMapper::New();
    mapWindow2->SetInput(window2Geom->GetOutput());
    mapWindow2->ScalarVisibilityOff();
  vtkActor *window2 = vtkActor::New();
    window2->SetMapper(mapWindow2);
    window2->GetProperty()->SetColor(.3,.3,.5);
    window2->SetUserMatrix(caveTransform->GetMatrixPointer());

  vtkStructuredGridGeometryFilter *klower1Geom = vtkStructuredGridGeometryFilter::New();
    klower1Geom->SetInput(reader->GetOutput());
    klower1Geom->SetExtent(17,17,0,11,0,6);
  vtkPolyDataMapper *mapKlower1 = vtkPolyDataMapper::New();
    mapKlower1->SetInput(klower1Geom->GetOutput());
    mapKlower1->ScalarVisibilityOff();
  vtkActor *klower1 = vtkActor::New();
    klower1->SetMapper(mapKlower1);
    klower1->GetProperty()->SetColor(.8,.8,.6);
    klower1->SetUserMatrix(caveTransform->GetMatrixPointer());

  vtkStructuredGridGeometryFilter *klower2Geom = vtkStructuredGridGeometryFilter::New();
    klower2Geom->SetInput(reader->GetOutput());
    klower2Geom->SetExtent(19,19,0,11,0,6);
  vtkPolyDataMapper *mapKlower2 = vtkPolyDataMapper::New();
    mapKlower2->SetInput(klower2Geom->GetOutput());
    mapKlower2->ScalarVisibilityOff();
  vtkActor *klower2 = vtkActor::New();
    klower2->SetMapper(mapKlower2);
    klower2->GetProperty()->SetColor(.8,.8,.6);
    klower2->SetUserMatrix(caveTransform->GetMatrixPointer());

  vtkStructuredGridGeometryFilter *klower3Geom = vtkStructuredGridGeometryFilter::New();
    klower3Geom->SetInput(reader->GetOutput());
    klower3Geom->SetExtent(17,19,0,0,0,6);
  vtkPolyDataMapper *mapKlower3 = vtkPolyDataMapper::New();
    mapKlower3->SetInput(klower3Geom->GetOutput());
    mapKlower3->ScalarVisibilityOff();
  vtkActor *klower3 = vtkActor::New();
    klower3->SetMapper(mapKlower3);
    klower3->GetProperty()->SetColor(.8,.8,.6);
    klower3->SetUserMatrix(caveTransform->GetMatrixPointer());

  vtkStructuredGridGeometryFilter *klower4Geom = vtkStructuredGridGeometryFilter::New();
    klower4Geom->SetInput(reader->GetOutput());
    klower4Geom->SetExtent(17,19,11,11,0,6);
  vtkPolyDataMapper *mapKlower4 = vtkPolyDataMapper::New();
    mapKlower4->SetInput(klower4Geom->GetOutput());
    mapKlower4->ScalarVisibilityOff();
  vtkActor *klower4 = vtkActor::New();
    klower4->SetMapper(mapKlower4);
    klower4->GetProperty()->SetColor(.8,.8,.6);
    klower4->SetUserMatrix(caveTransform->GetMatrixPointer());


  vtkStructuredGridGeometryFilter *klower5Geom = vtkStructuredGridGeometryFilter::New();
    klower5Geom->SetInput(reader->GetOutput());
    klower5Geom->SetExtent(17,19,0,11,0,0);
  vtkPolyDataMapper *mapKlower5 = vtkPolyDataMapper::New();
    mapKlower5->SetInput(klower5Geom->GetOutput());
    mapKlower5->ScalarVisibilityOff();
  vtkActor *klower5 = vtkActor::New();
    klower5->SetMapper(mapKlower5);
    klower5->GetProperty()->SetColor(.8,.8,.6);
    klower5->SetUserMatrix(caveTransform->GetMatrixPointer());

  vtkStructuredGridGeometryFilter *klower6Geom = vtkStructuredGridGeometryFilter::New();
    klower6Geom->SetInput(reader->GetOutput());
    klower6Geom->SetExtent(17,19,0,7,6,6);
  vtkPolyDataMapper *mapKlower6 = vtkPolyDataMapper::New();
    mapKlower6->SetInput(klower6Geom->GetOutput());
    mapKlower6->ScalarVisibilityOff();
  vtkActor *klower6 = vtkActor::New();
    klower6->SetMapper(mapKlower6);
    klower6->GetProperty()->SetColor(.8,.8,.6);
    klower6->SetUserMatrix(caveTransform->GetMatrixPointer());

  vtkStructuredGridGeometryFilter *klower7Geom = vtkStructuredGridGeometryFilter::New();
    klower7Geom->SetInput(reader->GetOutput());
    klower7Geom->SetExtent(17,19,9,11,6,6);
  vtkPolyDataMapper *mapKlower7 = vtkPolyDataMapper::New();
    mapKlower7->SetInput(klower7Geom->GetOutput());
    mapKlower7->ScalarVisibilityOff();
  vtkActor *klower7 = vtkActor::New();
    klower7->SetMapper(mapKlower7);
    klower7->GetProperty()->SetColor(.8,.8,.6);
    klower7->SetUserMatrix(caveTransform->GetMatrixPointer());

  vtkStructuredGridGeometryFilter *hood1Geom = vtkStructuredGridGeometryFilter::New();
    hood1Geom->SetInput(reader->GetOutput());
    hood1Geom->SetExtent(17,17,0,11,11,16);
  vtkPolyDataMapper *mapHood1 = vtkPolyDataMapper::New();
    mapHood1->SetInput(hood1Geom->GetOutput());
    mapHood1->ScalarVisibilityOff();
  vtkActor *hood1 = vtkActor::New();
    hood1->SetMapper(mapHood1);
    hood1->GetProperty()->SetColor(.8,.8,.6);
    hood1->SetUserMatrix(caveTransform->GetMatrixPointer());

  vtkStructuredGridGeometryFilter *hood2Geom = vtkStructuredGridGeometryFilter::New();
    hood2Geom->SetInput(reader->GetOutput());
    hood2Geom->SetExtent(19,19,0,11,11,16);
  vtkPolyDataMapper *mapHood2 = vtkPolyDataMapper::New();
    mapHood2->SetInput(hood2Geom->GetOutput());
    mapHood2->ScalarVisibilityOff();
  vtkActor *hood2 = vtkActor::New();
    hood2->SetMapper(mapHood2);
    hood2->GetProperty()->SetColor(.8,.8,.6);
    hood2->SetUserMatrix(caveTransform->GetMatrixPointer());

  vtkStructuredGridGeometryFilter *hood3Geom = vtkStructuredGridGeometryFilter::New();
    hood3Geom->SetInput(reader->GetOutput());
    hood3Geom->SetExtent(17,19,0,0,11,16);
  vtkPolyDataMapper *mapHood3 = vtkPolyDataMapper::New();
    mapHood3->SetInput(hood3Geom->GetOutput());
    mapHood3->ScalarVisibilityOff();
  vtkActor *hood3 = vtkActor::New();
    hood3->SetMapper(mapHood3);
    hood3->GetProperty()->SetColor(.8,.8,.6);
    hood3->SetUserMatrix(caveTransform->GetMatrixPointer());


  vtkStructuredGridGeometryFilter *hood4Geom = vtkStructuredGridGeometryFilter::New();
    hood4Geom->SetInput(reader->GetOutput());
    hood4Geom->SetExtent(17,19,11,11,11,16);
  vtkPolyDataMapper *mapHood4 = vtkPolyDataMapper::New();
    mapHood4->SetInput(hood4Geom->GetOutput());
    mapHood4->ScalarVisibilityOff();
  vtkActor *hood4 = vtkActor::New();
    hood4->SetMapper(mapHood4);
    hood4->GetProperty()->SetColor(.8,.8,.6);
    hood4->SetUserMatrix(caveTransform->GetMatrixPointer());

  vtkStructuredGridGeometryFilter *hood6Geom = vtkStructuredGridGeometryFilter::New();
    hood6Geom->SetInput(reader->GetOutput());
    hood6Geom->SetExtent(17,19,0,11,16,16);
  vtkPolyDataMapper *mapHood6 = vtkPolyDataMapper::New();
    mapHood6->SetInput(hood6Geom->GetOutput());
    mapHood6->ScalarVisibilityOff();
  vtkActor *hood6 = vtkActor::New();
    hood6->SetMapper(mapHood6);
    hood6->GetProperty()->SetColor(.8,.8,.6);
    hood6->SetUserMatrix(caveTransform->GetMatrixPointer());

  vtkStructuredGridGeometryFilter *cookingPlateGeom = vtkStructuredGridGeometryFilter::New();
    cookingPlateGeom->SetInput(reader->GetOutput());
    cookingPlateGeom->SetExtent(17,19,7,9,6,6);
  vtkPolyDataMapper *mapCookingPlate = vtkPolyDataMapper::New();
    mapCookingPlate->SetInput(cookingPlateGeom->GetOutput());
    mapCookingPlate->ScalarVisibilityOff();
  vtkActor *cookingPlate = vtkActor::New();
    cookingPlate->SetMapper(mapCookingPlate);
    cookingPlate->GetProperty()->SetColor(.9,.1,.1);
    cookingPlate->SetUserMatrix(caveTransform->GetMatrixPointer());

  vtkStructuredGridGeometryFilter *filterGeom = vtkStructuredGridGeometryFilter::New();
    filterGeom->SetInput(reader->GetOutput());
    filterGeom->SetExtent(17,19,7,9,11,11);
  vtkPolyDataMapper *mapFilter = vtkPolyDataMapper::New();
    mapFilter->SetInput(filterGeom->GetOutput());
    mapFilter->ScalarVisibilityOff();
  vtkActor *filter = vtkActor::New();
    filter->SetMapper(mapFilter);
    filter->GetProperty()->SetColor(.8,.6,.6);
    filter->SetUserMatrix(caveTransform->GetMatrixPointer());


//
// This section diverges from the vtk example more sigificantly
// create the streamlines
//
  // rect, rake, and lines are global variable so that the frame 
  // function can update them.
  // changed the source from line to plane to make picking and
  // movement easier.  Line could be used without much trouble
  // a tube would need to be placed on it to make it big enough
  // to pick it.
  rect = vtkPlaneSource::New();
    rect->SetOrigin(0.08, 3.40, 0.61);
    rect->SetPoint1(0.08, 3.40, 0.81);
    rect->SetPoint2(0.08, 3.60, 0.61);
    rect->SetResolution(1, 1);
  vtkPolyDataMapper *rakeMapper = vtkPolyDataMapper::New();
    rakeMapper->SetInput(rect->GetOutput());
  rake = vtkActor::New();
    rake->SetMapper(rakeMapper);
    rake->SetUserMatrix(caveTransform->GetMatrixPointer());

  vtkStreamLine *streamers = vtkStreamLine::New();
    streamers->SetInput(reader->GetOutput());
    streamers->SetSource(rect->GetOutput());
    streamers->SetMaximumPropagationTime(maxTime);
    streamers->SetStepLength(maxTime/500.0);
    streamers->SetIntegrationStepLength(0.2);
    streamers->Update();
  vtkTubeFilter *streamTubes = vtkTubeFilter::New();
    streamTubes->SetInput(streamers->GetOutput());
    streamTubes->SetNumberOfSides(8);
    streamTubes->SetRadius(.02);
    streamTubes->SetVaryRadius(0);
  vtkPolyDataMapper *streamersMapper = vtkPolyDataMapper::New();
    streamersMapper->SetInput(streamTubes->GetOutput());
    streamersMapper->SetScalarRange(range);
//    streamersMapper->ScalarVisibilityOff();
  lines = vtkActor::New();
    lines->SetMapper(streamersMapper);
    lines->GetProperty()->SetColor(0,0,0);
    lines->SetUserMatrix(caveTransform->GetMatrixPointer());


    // Add the actors to the cave renderer
    caveRenderer->AddActor(outline);   
    caveRenderer->AddActor(door);   
    caveRenderer->AddActor(window1);   
    caveRenderer->AddActor(window2);   
    caveRenderer->AddActor(klower1);   
    caveRenderer->AddActor(klower2);   
    caveRenderer->AddActor(klower3);   
    caveRenderer->AddActor(klower4);   
    caveRenderer->AddActor(klower5);   
    caveRenderer->AddActor(klower6);   
    caveRenderer->AddActor(klower7);   
    caveRenderer->AddActor(hood1);   
    caveRenderer->AddActor(hood2);   
    caveRenderer->AddActor(hood3);   
    caveRenderer->AddActor(hood4);   
    caveRenderer->AddActor(hood6);   
    caveRenderer->AddActor(cookingPlate);   
    caveRenderer->AddActor(filter);   
    caveRenderer->AddActor(lines);   
    caveRenderer->AddActor(rake);   

    // add our only item to the pick list
    cavePicker->AddPickList(rake);

}

void
main(int argc, char **argv)
{
    CAVEConfigure(&argc,argv,NULL);
    CAVEInit();
    CAVEInitApplication((CAVECALLBACK)initVTK, 2, argc, argv);
    CAVEFrameFunction(frameUpdate, 0);
    CAVEDisplay(draw, 0);
    
    while(!CAVEgetbutton(CAVE_ESCKEY))
    {
    }
    
    CAVEExit();

    caveRenderer->Delete();
    caveRenderWindow->Delete();
    cavePicker->Delete();
}

// From the wands position and orientation determine
// its transformation matrix.
void
getWandMatrix(vtkMatrix4x4* m)
{
    vtkTransform *transform;
    float or[3],pos[3];

    transform = vtkTransform::New();

    CAVEGetPosition(CAVE_WAND_NAV,pos);
    CAVEGetOrientation(CAVE_WAND_NAV,or);

    // note the order of rotation.  This is very important
    transform->Translate(pos[0], pos[1], pos[2]);
    transform->RotateY(or[1]);
    transform->RotateX(or[0]);
    transform->RotateZ(or[2]);

    m->DeepCopy(transform->GetMatrixPointer());
    transform->Delete();
}

// Handles the translation component of navigation
// based on the orientation of the wand
void
navTranslate(float dz)
{

    float wandOrn[3];
    float wx, wz;

    CAVEGetOrientation(CAVE_WAND,wandOrn);
    wx = dz * sinf(DTOR(wandOrn[1]))*5;
    wz = dz * cosf(DTOR(wandOrn[1]))*5;
    CAVENavTranslate(wx, 0, wz);
}


// Handles the rotation component of navigation
// based on the amount the joystick is pressed
void
CAVENavRotateY()
{
    float azi;

    azi = 1.0;

    if(CAVE_JOYSTICK_X > 0.5)
        CAVENavRot(-azi, 'y');

    if(CAVE_JOYSTICK_X < -0.5)
        CAVENavRot(azi, 'y');
}


// The basic navigation function. It calls the
// auxillary navigation functions.
void
navigate()
{

    float dz;

    if(CAVE_JOYSTICK_Y > 0.5)
    {
        dz = -.1;
        navTranslate(dz);
    }
    if(CAVE_JOYSTICK_Y < -0.5)
    {
        dz = .1;
        navTranslate(dz);
    }
    CAVENavRotateY();
}

