/* vtkCAVEPicker - Extends vtkPicker to work within CAVE environment
Inherits from vtkPicker, requires vtk library available from kitware
inc. <www.kitware.com>

This software has been developed as part of the Futures Laboratory
project at Argonne National Laboratory, for more information email
fl-info@mcs.anl.gov, and is distributed under the terms described at
http://www.mcs.anl.gov/fl/softwareterms.html

(c) 1998 Argonne National Laboratory/University of Chicago

*/
#include <vtkCAVEPicker.h>
#include <vtkMath.h>

int
vtkCAVEPicker::Pick(vtkRenderer *renderer, float* origin,
		    float * vector, float front, float back)
{
    int i;
    float t;
    vtkActorCollection *actors;
    vtkActor *actor, *part;
    vtkMapper *mapper;
    float p1World[4], p2World[4], p1Mapper[4], p2Mapper[4];
    int picked=0;
    float ray[3];
    float opacity;
    int visible, pickable;
    float *bounds, tol;
    float tF, tB;
    float hitPosition[3];
  
    // Invoke start pick method if defined
    if ( this->StartPickMethod )
	(*this->StartPickMethod)(this->StartPickMethodArg);

    //  Initialize picking process
    this->Renderer = renderer;

    this->Initialize();

    if ( renderer == NULL )
    {
	vtkErrorMacro(<<"Must specify renderer!");
	return 0;
    }

    //  Compute the ray endpoints.  The ray is along the line running from
    //  the camera position to the selection point, starting where this line
    //  intersects the front clipping plane, and terminating where this
    //  line intersects the back clipping plane.

    tF = front;
    tB = back;

    for (i=0; i<3; i++)
    {
	p1World[i] = origin[i] + tF*vector[i];
	p2World[i] = origin[i] + tB*vector[i];
    }
    p1World[3] = 1.0;
    p2World[3] = 1.0;
	

    // tol does NOT get used by IntersectWithLine(), so the code to calculate
    // a tolerance has been removed, and tol is set to an arbitrary value.
    tol = 1;

    // Loop over all actors.  Transform ray (defined from user)
    // into coordinates of mapper (not transformed to actors coordinates!
    // Reduces overall computation!!!).
    //
    if ( this->PickFromList )
	actors = this->GetPickList();
    else
	actors = renderer->GetActors();

    this->Transform->PostMultiply();

    for ( actors->InitTraversal(); (actor=actors->GetNextItem()); )
    {
	for ( actor->InitPartTraversal(); (part=actor->GetNextPart()); )
	{
	    visible = part->GetVisibility();
	    pickable = part->GetPickable();
	    opacity = part->GetProperty()->GetOpacity();

	    //  If actor can be picked, get its composite matrix, invert it, and
	    //  use the inverted matrix to transform the ray points into mapper
	    //  coordinates. 
	    if ( visible && pickable && (opacity != 0.0) &&
		 (mapper = part->GetMapper()) != NULL )
	    {
		// Get the current actor's transformation matrix. Inverse this
		// matrix and apply it to the ray's starting and ending positions

		this->Transform->SetMatrix(part->vtkProp3D::GetMatrix());
		this->Transform->Push();

		this->Transform->Inverse();
		
		this->Transform->SetPoint(p1World);
		this->Transform->GetPoint(p1Mapper);

		this->Transform->SetPoint(p2World);
		this->Transform->GetPoint(p2Mapper);

		this->Transform->Pop();
		
		// Calculate the intersection ray based on starting and ending
		// points of the ray. Do NOT normalize the ray, the ray's length
		// determines the length of allowable intersection testing.
		for (i=0; i<3; i++) 
		    ray[i] = p2Mapper[i] - p1Mapper[i];

		//  Get the bounding box of the modeller.  Note that the tolerance is
		//  added to the bounding box to make sure things on the edge of the
		//  bounding box are picked correctly.

		bounds = mapper->GetBounds();

		if ( vtkCell::HitBBox(bounds, (float *)p1Mapper, ray,
				      hitPosition, t) )
		{
		    picked = 1;
		    this->IntersectWithLine((float *)p1Mapper,
					    (float *)p2Mapper,tol,actor,part,
					    mapper);
		    //this->Actors.AddItem(part);
		}

	    }//if visible and pickable not transparent and has mapper
	}//for all parts
    }//for all actors

    // Invoke end pick method if defined
    if ( this->EndPickMethod )
	(*this->EndPickMethod)(this->EndPickMethodArg);

    return picked;
}

// Intersect data with specified ray.
float
vtkCAVEPicker::IntersectWithLine(float p1[3], float p2[3], 
				 float vtkNotUsed(tol), vtkActor *assem, 
				 vtkActor *actor, vtkMapper *mapper)
{
    int i;
    float *center, t, ray[3], rayFactor;


    //
    // Get the data from the modeller
    //
    center = mapper->GetCenter();

    for (i=0; i<3; i++)
	ray[i] = p2[i] - p1[i];

    if (( rayFactor = vtkMath::Dot(ray,ray)) == 0.0 )
    	return 2.0; // Changed 1/5/99 to mirror vtk changes
    
    //
    // Project the center point onto the ray and determine its parametric value
    //
    t = (ray[0]*(center[0]-p1[0]) + ray[1]*(center[1]-p1[1])
	 + ray[2]*(center[2]-p1[2])) / rayFactor;

    // Had to change the range of t to be from (-1,1) instead of (0,1)
    // The (0,1) range may have had to do with the fact the original
    // picker was designed for mouse picking in 2D windows. ???
    if ( t >= -1.0 && t <= 1.0 && t < this->GlobalTMin ) 
	this->MarkPicked(assem, actor, mapper, t, center);

    return t; // Changed 1/5/99 to mirror vtk changes
}
